# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Low-level communication with the Transmission daemon"""

import asyncio
import json

import async_timeout
from blinker import Signal

from ..errors import AuthError, ClientError, ConnectionError, RPCError, TimeoutError
from ..utils import URL

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


AUTH_ERROR_CODE = 401
CSRF_ERROR_CODE = 409
CSRF_HEADER = 'X-Transmission-Session-Id'
TIMEOUT = 10


class TransmissionRPC():
    """
    Low-level AsyncIO Transmission RPC communication

    This class handles connecting to a Transmission daemon via the RPC
    interface.  It does not implement the RPC protocol, only basic things like
    authentication, sending requests and receiving responses.  High-level RPCs
    are done in the *API classes.
    """

    def __init__(self, host='localhost', port=9091, *, tls=False, user='',
                 password='', proxy='', path='/transmission/rpc', enabled=True):
        self.host = host
        self.port = port
        self.path = path
        self.tls = tls
        self.user = user
        self.password = password
        self.proxy = proxy
        self._headers = {'content-type': 'application/json'}
        self._session = None
        self._enabled_event = asyncio.Event()
        self.enabled = enabled
        self._request_lock = asyncio.Lock()
        self._connecting_lock = asyncio.Lock()
        self._connection_tested = False
        self._connection_exception = None
        self._timeout = TIMEOUT
        self._version = None
        self._rpcversion = None
        self._rpcversionmin = None
        self._on_connecting = Signal()
        self._on_connected = Signal()
        self._on_disconnected = Signal()
        self._on_error = Signal()

    def on(self, signal, callback, autoremove=True):
        """
        Register `callback` for `signal`

        signal: 'connecting', 'connected', 'disconnected' or 'error'
        callback: a callable that receives this instance as a positional
                  argument and, in case of the 'error' signal, the exception as
                  a keyword argument with the name 'error'

        Callbacks are automatically unsubscribed when they are
        garbage-collected.
        """
        try:
            sig = getattr(self, '_on_' + signal)
        except AttributeError:
            raise ValueError('Unknown signal: {!r}'.format(signal))
        else:
            if not isinstance(sig, Signal):
                raise ValueError('Unknown signal: {!r}'.format(signal))
            else:
                log.debug('Registering %r for %r event', callback, signal)
                sig.connect(callback, weak=autoremove)

    @property
    def version(self):
        """Version of the Transmission daemon or None if not connected"""
        return self._version

    @property
    def rpcversion(self):
        """RPC version of the Transmission daemon or None if not connected"""
        return self._rpcversion

    @property
    def rpcversionmin(self):
        """Oldest RPC version supported by Transmission daemon or None if not connected"""
        return self._rpcversionmin

    @property
    def host(self):
        """
        Hostname or IP of the Transmission RPC interface

        Setting this property calls disconnect().
        """
        return self._host

    @host.setter
    def host(self, host):
        self._host = str(host) if host is not None else 'localhost'
        asyncio.ensure_future(self.disconnect('Changing host: %r' % self._host))

    @property
    def path(self):
        """
        Path of the Transmission RPC interface

        Setting this property calls disconnect().
        """
        return self._path

    @path.setter
    def path(self, path):
        if path is None:
            path = '/transmission/rpc'
        elif not path or path[0] != '/':
            path = '/' + path
        self._path = path
        asyncio.ensure_future(self.disconnect('Changing path: %r' % self._path))

    @property
    def port(self):
        """
        Port of the Transmission RPC interface

        Setting this property calls disconnect().
        """
        return self._port

    @port.setter
    def port(self, port):
        self._port = int(port) if port is not None else 9091
        asyncio.ensure_future(self.disconnect('Changing port: %r' % self._port))

    @property
    def user(self):
        """
        Username for authenticating to the Transmission RPC interface or empty string

        Setting this property calls disconnect().
        """
        return self._user

    @user.setter
    def user(self, user):
        self._user = str(user) if user is not None else ''
        asyncio.ensure_future(self.disconnect('Changing user: %r' % self._user))

    @property
    def password(self):
        """
        Password for authenticating to the Transmission RPC interface or empty string

        Setting this property calls disconnect().
        """
        return self._password

    @password.setter
    def password(self, password):
        self._password = str(password) if password is not None else ''
        asyncio.ensure_future(self.disconnect('Changing password: %r' % self._password))

    @property
    def tls(self):
        """
        Whether to use HTTPS for connecting to the Transmission RPC interface

        Setting this property calls disconnect().
        """
        return self._tls

    @tls.setter
    def tls(self, tls):
        self._tls = bool(tls) if tls is not None else False
        asyncio.ensure_future(self.disconnect('Changing tls: %r' % self._tls))

    @property
    def url(self):
        """
        URL of the Transmission RPC interface

        Setting or getting this property sets or gets the following properties: tls, user,
        password, host, port, path

        Missing parts are filled in with defaults, e.g. "example.org:1234" results in
        "http://example.org:1234/transmission/rpc".

        While user and password are supported when setting ("user:password@localhost"),
        they are not included in the return value for security reasons.  See the
        url_unsafe property.
        """
        return '%s://%s:%d%s' % ('https' if self.tls else 'http',
                                 self.host, self.port, self.path)

    @url.setter
    def url(self, url):
        if url is None:
            url = URL('http://localhost:9091/transmission/rpc')
        else:
            url = URL(url)
        self._user = url.user or ''
        self._password = url.password or ''
        if url.scheme == 'https':
            self._tls = True
        elif url.scheme == 'http':
            self._tls = False
        else:
            raise ValueError('Invalid scheme: %r' % (url.scheme,))
        self._host = url.host
        self._port = int(url.port) if url.port is not None else 9091
        self._path = url.path if url.path is not None else '/transmission/rpc'
        asyncio.ensure_future(self.disconnect('Changing url: %r' % self._url))

    @property
    def url_unsafe(self):
        """URL of the Transmission RPC interface with user and password if given"""
        if self.user or self.password:
            return '%s://%s:%s@%s:%d%s' % ('https' if self.tls else 'http', self.user,
                                           self.password, self.host, self.port, self.path)
        else:
            return self.url

    @property
    def proxy(self):
        """URL of a SOCKS5, SOCKS4 or HTTP proxy"""
        return self._proxy

    @proxy.setter
    def proxy(self, proxy):
        if proxy:
            try:
                import aiohttp_socks
            except ImportError:
                raise ValueError('Missing extra: proxy (aiohttp_socks)')
            self._connector = aiohttp_socks.ProxyConnector.from_url(proxy)
            self._proxy = URL(proxy)
        else:
            self._connector = None
            self._proxy = URL('')
        asyncio.ensure_future(self.disconnect('Changing proxy: %r' % self._proxy))

    @property
    def timeout(self):
        """Number of seconds to try to connect before giving up"""
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        self._timeout = float(timeout)

    @property
    def enabled(self):
        """
        Whether requests should connect

        If this is set to False, requests will wait for it to be set to True.
        This allows you to block any connection attempts until the connection
        parameters (host, user, password, etc) are specified to prevent any
        unwarranted error messages.
        """
        return self._enabled_event.is_set()

    @enabled.setter
    def enabled(self, enabled):
        if enabled and not self.enabled:
            log.debug('Enabling %r', self)
            self._enabled_event.set()
        elif not enabled and self.enabled:
            log.debug('Disabling %r', self)
            self._enabled_event.clear()
            if self.connected:
                asyncio.ensure_future(self.disconnect())

    @property
    def connected(self):
        """Return True if connected, False otherwise"""
        return (self._session is not None
                and not self._session.closed
                and self._connection_tested)

    async def connect(self):
        """
        Connect to running daemon

        If the `enabled` property is set to False, this method blocks until
        `enabled` is set to True.

        Raises RPCError, ConnectionError or AuthError.
        """
        log.debug('Connecting to %s (timeout=%ss)', self.url, self.timeout)
        self._on_connecting.send(self)

        if self._connecting_lock.locked():
            if self._connection_exception is not None:
                # The other connect() call failed
                log.debug('Found connection error: %r', self._connection_exception)
                raise self._connection_exception

            log.debug('Connection is already being established - Waiting ...')
            try:
                async with async_timeout.timeout(self.timeout):
                    await self._enabled_event.wait()
            except asyncio.TimeoutError:
                raise TimeoutError(self.timeout, self.url)
            else:
                if self.connected:
                    log.debug('Connection is up: %r', self.url)
                    return

        async with self._connecting_lock:
            log.debug('Acquired connect() lock')

            if self.connected:
                await self.disconnect('Reconnecting')

            # Block until we're enabled
            await self._enabled_event.wait()

            import aiohttp
            session_args = {}
            if self.user or self.password:
                session_args['auth'] = aiohttp.BasicAuth(self.user, self.password,
                                                         encoding='utf-8')
            if self._connector is not None:
                session_args['connector'] = self._connector
                session_args['connector_owner'] = False
            self._session = aiohttp.ClientSession(**session_args)

            # Check if connection works
            log.debug('Testing connection to %s', self.url)
            try:
                test_request = json.dumps({'method':'session-get'})
                info = await self._send_request(test_request)
            except ClientError as e:
                self._connection_exception = e
                log.debug('Caught during connection test: %r', e)
                await self._reset()
                self._on_error.send(self, error=e)
                raise
            else:
                self._version = info['version']
                self._rpcversion = info['rpc-version']
                self._rpcversionmin = info['rpc-version-minimum']
                self._connection_tested = True
                self._connection_exception = None
                log.debug('Connection established: %s', self.url)
                self._on_connected.send(self)

            log.debug('Releasing connect() lock')

    async def disconnect(self, reason=None):
        """
        Disconnect if connected

        reason: Why are we disconnecting? Only used in a debugging message.
        """
        if self.connected:
            await self._reset()
            log.debug('Disconnecting from %s (%s)', self.url,
                      reason if reason is not None else 'for no reason')
            self._on_disconnected.send(self)

    async def _reset(self):
        if self._session is not None:
            await self._session.close()
        self._session = None
        self._version = None
        self._rpcversion = None
        self._rpcversionmin = None
        self._connection_tested = False

    async def _post(self, data):
        async with async_timeout.timeout(self.timeout):
            response = await self._session.post(self.url, data=data, headers=self._headers)

            if response.status == CSRF_ERROR_CODE:
                # Send request again with CSRF header
                self._headers[CSRF_HEADER] = response.headers[CSRF_HEADER]
                log.debug('Setting CSRF header: %s = %s',
                          CSRF_HEADER, response.headers[CSRF_HEADER])
                await response.release()
                return await self._post(data)

            elif response.status == AUTH_ERROR_CODE:
                await response.release()
                log.debug('Authentication failed: %s: user=%r, password=%r',
                          self.url, self.user, self.password)
                raise AuthError(self.url)

            else:
                try:
                    answer = await response.json()
                except json.JSONDecodeError as e:
                    raise RPCError('Server sent malformed JSON: %s: %s' % (e, await response.text()))
                except UnicodeDecodeError as e:
                    raise RPCError('Server sent malformed UTF-8: %s' % e)
                else:
                    return answer

    async def _send_request(self, post_data):
        """
        Send RPC POST request to daemon

        post_data: Any valid RPC request as JSON string

        If applicable, returns response['arguments']['torrents'] or
        response['arguments'], otherwise response.

        Raises ClientError.
        """
        # NOTE #163: Letting asyncio.CancelledError bubble up seems to fix the issue that
        #            causes empty torrent lists in new tabs until the next poll iteration.
        #            But I've seen this error pop up in the TUI log: "Unclosed client
        #            session client_session: <aiohttp.client.ClientSession object at
        #            0x7f35d98d1be0>" This may or may not be related.
        import aiohttp
        try:
            from aiohttp_socks import ProxyConnectionError, ProxyError, ProxyTimeoutError
        except ImportError:
            class ProxyError(Exception): pass
            class ProxyConnectionError(Exception): pass
            class ProxyTimeoutError(Exception): pass
        try:
            answer = await self._post(post_data)
        except aiohttp.ClientError as e:
            log.debug('Caught during POST request: %r', e)
            raise ConnectionError(self.url)

        except (ProxyError,ProxyConnectionError) as e:
            log.debug('Caught during POST request: %r', e)
            raise ConnectionError(self.proxy)

        except asyncio.TimeoutError as e:
            log.debug('Caught during POST request: %r', e)
            raise TimeoutError(self.timeout, self.url)

        except ProxyTimeoutError as e:
            log.debug('Caught during POST request: %r', e)
            raise TimeoutError(self.timeout, self.proxy)

        else:
            if answer['result'] != 'success':
                raise RPCError(answer['result'].capitalize())
            else:
                if 'arguments' in answer:
                    if 'torrents' in answer['arguments']:
                        return answer['arguments']['torrents']
                    else:
                        return answer['arguments']
                return answer

    def __getattr__(self, method):
        """
        Return asyncio coroutine that sends RPC request and returns response

        method: Any method from the RPC specs with every '-' replaced with '_'.
                For arguments see the RPC specs.

        Example:
        >>> stats = await client.session_stats()
        >>> torrents = await client.torrent_get(ids=(1,2,3), fields=('status','name'))

        Raises RPCError, ConnectionError, AuthError
        """
        async def request(arguments=None, **kwargs):
            arguments = arguments or {}

            async with self._request_lock:
                if not self.connected:
                    log.debug('Autoconnecting for %r', method)
                    await self.connect()

                arguments.update(**kwargs)
                data = {'method'    : method.replace('_', '-'),
                        'arguments' : arguments}
                try:
                    rpc_request = json.dumps(data)
                except Exception as e:
                    raise RuntimeError('Invalid JSON data: %s: %r' % (e, data)) from None

                try:
                    return await self._send_request(rpc_request)
                except ClientError as e:
                    log.debug('Caught ClientError in %r request: %r', method, e)

                    # RPCError does not mean host is unreachable, there was just a
                    # misunderstanding, so we're still connected.
                    if not isinstance(e, RPCError) and self.connected:
                        await self.disconnect(str(e))

                    self._on_error.send(self, error=e)
                    raise

        request.__name__ = method
        request.__qualname__ = method
        return request



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import os
from collections import abc, defaultdict
from types import SimpleNamespace

import blinker

from ..poll import RequestPoller
from ..utils import Bool, BoolOrBandwidth, BoolOrPath, Int, Option, Path, const, convert

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


# Transform key (as in `settings[key]`) to property name and vice versa
def _key2property(key):
    return key.replace('.', '_')
def _property2key(property_name):
    return property_name.replace('_', '.')


_SETTINGS = {}
def _setting(value_cls, **kwargs):
    """Decorator for SettingsAPI properties"""
    def wrap(method):
        property_name = method.__name__
        setting_name = _property2key(property_name)
        description = kwargs.pop('description') if 'description' in kwargs else None
        converter = value_cls.partial(**kwargs)

        def get_typespec(api):
            return SimpleNamespace(
                converter=converter,
                description=description or getattr(type(api), property_name).__doc__)
        _SETTINGS[setting_name] = get_typespec

        return property(method)
    return wrap


class SettingsAPI(abc.Mapping, RequestPoller):
    """
    Transmission daemon settings

    `set_*` methods are coroutine functions that request value changes from the
    server.

    `get_*` methods are coroutine functions that fetch values from the server.

    Cached values, which are updated every `interval` seconds, are available as
    properties with 'get_' removed from the equivalent method name
    (e.g. api.get_port() -> api.port).

    Use `on_update` to set a callback for interval updates.

    Cached values are also available as mapping items with '.' instead of '_',
    e.g.  `settings['path.incomplete']`.

    To update cached values, use `update` (async) or `poll` (sync).
    """

    # Mapping methods
    def __getitem__(self, key):
        try:
            item = getattr(self, _key2property(key))
        except AttributeError:
            raise KeyError(key)
        else:
            return item

    def __contains__(self, key):
        return key in self._cache

    def __iter__(self):
        return iter(self._cache)

    def __len__(self):
        return len(self._cache)


    def __init__(self, srvapi, interval=1):
        self._cache = {}
        self._descriptions = {}
        self._converters = {}
        for setting,get_typespec in _SETTINGS.items():
            typespec = get_typespec(self)
            self._cache[setting] = None
            self._converters[setting] = typespec.converter
            self._descriptions[setting] = typespec.description

        self._raw = None    # Raw dict from 'session-get' or None if not connected
        self._srvapi = srvapi
        self._on_update = blinker.Signal()
        self._on_set = defaultdict(lambda: blinker.Signal())

        super().__init__(self._srvapi.rpc.session_get, interval=interval)
        self.on_response(self._handle_session_get)
        self.on_error(self._handle_error)

    def description(self, name):
        """Return setting's description"""
        return self._descriptions[name]

    def syntax(self, name):
        """Return setting's value syntax"""
        return self._converters[name].syntax

    def default(self, name):
        # Maintain consistency with local Settings class
        return const.DISCONNECTED

    def validate(self, name, value):
        if not isinstance(value, str) and isinstance(value, abc.Iterable):
            return self._converters[name](*value)
        else:
            return self._converters[name](value)

    @property
    def as_dict(self):
        # NOTE 1: The "id" key is important because filters (including
        #         SettingFilter) expect it.  This requirement can possibly
        #         removed but it probably means rewriting lots of filter tests.
        # NOTE 2: The "validate" lambda MUST store the value of `name`,
        #         otherwise `self.validate` always gets called with the for
        #         loop's last `name` value.
        return {name:{'id': name,
                      'value': self[name],
                      'default': self.default(name),
                      'description': self.description(name),
                      'syntax': self.syntax(name),
                      'validate': lambda v, n=name: self.validate(n, v)}
                for name,value in self.items()}

    async def update(self):
        """Request update from server"""
        log.debug('Requesting immediate settings update')
        self._handle_session_get(await self.request())

    def _handle_session_get(self, response):
        """Request update from server"""
        log.debug('Handling settings update')
        self.clearcache(run_callbacks=False)
        self._raw = response
        self._on_update.send(self)

    def _handle_error(self, error):
        self.clearcache(run_callbacks=True)

    def clearcache(self, run_callbacks=True):
        """
        Clear cached settings

        update: Whether to run `on_update` callbacks
        """
        log.debug('Clearing %s cache', type(self).__name__)
        self._raw = None
        for setting in self._cache:
            self._cache[setting] = None
        if run_callbacks:
            self._on_update.send(self)

    def on_update(self, callback, autoremove=True):
        """
        Register `callback` to be called when settings have changed

        `callback` gets the instance of this class.

        If `autoremove` is True, `callback` is removed automatically when it is
        deleted.
        """
        log.debug('Registering %r to receive settings updates', callback)
        self._on_update.connect(callback, weak=autoremove)

    def on_set(self, callback, key=None, autoremove=True):
        """
        Register `callback` to be called after settings are set

        `callback` gets the instance of this class.

        If `key` is None, `callback` is called for every setting.  If `key` is the name of
        a setting, `callback` is only called for that setting.

        If `autoremove` is True, `callback` is removed automatically when it is
        deleted.
        """
        log.debug('Registering %r to be called after setting changes: %r', callback, key)
        if key is not None and key not in self:
            raise ValueError(key)
        else:
            log.debug(self._on_set[key])
            log.debug('%r.connect(%r, weak=%r)', self._on_set[key], callback, autoremove)
            self._on_set[key].connect(callback, weak=autoremove)

    async def get(self, key):
        """Same as __getitem__ but refresh cache first"""
        if key not in self:
            raise ValueError(key)
        else:
            method = getattr(self, 'get_' + _key2property(key))
            return await method()

    async def set(self, key, value):
        """Asynchronous replacement for __setitem__"""
        if key not in self:
            raise ValueError(key)
        else:
            method = getattr(self, 'set_' + _key2property(key))
            await method(value)

    def _get(self, key, field_or_callable):
        """Get setting from cache if possible"""
        if self._cache[key] is None:
            if self._raw is None:
                self._cache[key] = const.DISCONNECTED
            elif callable(field_or_callable):
                self._cache[key] = self._converters[key](field_or_callable())
            else:
                self._cache[key] = self._converters[key](self._raw[field_or_callable])
        return self._cache[key]

    async def _set(self, key, request):
        """Send 'session-set' request with dictionary `request` and call `update`"""
        log.debug('Sending session-set request: %r', request)
        await self._srvapi.rpc.session_set(request)
        await self.update()
        self._on_set[key].send(self)
        self._on_set[None].send(self)


    @_setting(Bool)
    def autostart(self):
        """Whether added torrents are started automatically"""
        return self._get('autostart', 'start-added-torrents')

    async def get_autostart(self):
        """Refresh cache and return `autostart`"""
        await self.update()
        return self.autostart

    async def set_autostart(self, enabled):
        """See `autostart`"""
        value = self._converters['autostart'](enabled)
        await self._set('autostart', {'start-added-torrents': bool(value)})


    # Network settings

    @_setting(Int, min=0, max=65535, prefix='none')  # noqa: E301
    def port(self):
        """Port used to communicate with peers"""
        return self._get('port', 'peer-port')

    async def get_port(self):
        """Refresh cache and return `port`"""
        await self.update()
        return self.port

    async def set_port(self, port):
        """See `port`"""
        value = self._converters['port'](port)
        await self._set('port', {'peer-port': value,
                                 'peer-port-random-on-start': False})


    @_setting(Bool)
    def port_random(self):
        """Whether to pick a random port when the daemon starts"""
        return self._get('port.random', 'peer-port-random-on-start')

    async def get_port_random(self):
        """Refresh cache and return `port_random`"""
        await self.update()
        return self.port_random

    async def set_port_random(self, port_random):
        """See `port_random`"""
        value = self._converters['port.random'](port_random)
        await self._set('port.random', {'peer-port-random-on-start': bool(value)})


    @_setting(Bool)
    def port_forwarding(self):
        """Whether to autoconfigure port-forwarding via UPnP/NAT-PMP"""
        return self._get('port.forwarding', 'port-forwarding-enabled')

    async def get_port_forwarding(self):
        """Refresh cache and return `port_forwarding`"""
        await self.update()
        return self.port_forwarding

    async def set_port_forwarding(self, enabled):
        """See `port_forwarding`"""
        value = self._converters['port.forwarding'](enabled)
        await self._set('port.forwarding', {'port-forwarding-enabled': bool(value)})


    @_setting(Int, min=1, max=65535)
    def limit_peers_global(self):
        """Maximum number of connections for all torrents combined"""
        return self._get('limit.peers.global', 'peer-limit-global')

    async def get_limit_peers_global(self):
        """Refresh cache and return `limit_peers_global`"""
        await self.update()
        return self.limit_peers_global

    async def set_limit_peers_global(self, limit):
        """See `limit_peers_global`"""
        value = self._converters['limit.peers.global'](limit)
        await self._set('limit.peers.global', {'peer-limit-global': value})


    @_setting(Int, min=1, max=65535)
    def limit_peers_torrent(self):
        """Maximum number of connections per torrent"""
        return self._get('limit.peers.torrent', 'peer-limit-per-torrent')

    async def get_limit_peers_torrent(self):
        """Refresh cache and return `limit_peers_torrent`"""
        await self.update()
        return self.limit_peers_torrent

    async def set_limit_peers_torrent(self, limit):
        """See `limit_peers_torrent`"""
        value = self._converters['limit.peers.torrent'](limit)
        await self._set('limit.peers.torrent', {'peer-limit-per-torrent': value})


    @_setting(Option, options=('required', 'preferred', 'tolerated'),
              description='Protocol encryption policy; "required", "preferred" or "tolerated"')
    def encryption(self):
        """
        Whether protocol encryption is used to mask BitTorrent traffic

        One of the strings 'required', 'preferred' or 'tolerated'.
        """
        return self._get('encryption', 'encryption')

    async def get_encryption(self):
        """Refresh cache and return `encryption`"""
        await self.update()
        return self.encryption

    async def set_encryption(self, encryption):
        """See `encryption`"""
        value = self._converters['encryption'](encryption)
        await self._set('encryption', {'encryption': value})


    @_setting(Bool)
    def utp(self):
        """Whether to use µTP to mitigate latency issues"""
        return self._get('utp', 'utp-enabled')

    async def get_utp(self):
        """Refresh cache and return `utp`"""
        await self.update()
        return self.utp

    async def set_utp(self, enabled):
        """See `utp`"""
        value = self._converters['utp'](enabled)
        await self._set('utp', {'utp-enabled': bool(value)})


    @_setting(Bool)
    def dht(self):
        """Whether to use DHT to discover peers for public torrents"""
        return self._get('dht', 'dht-enabled')

    async def get_dht(self):
        """Refresh cache and return `dht`"""
        await self.update()
        return self.dht

    async def set_dht(self, enabled):
        """See `dht`"""
        value = self._converters['dht'](enabled)
        await self._set('dht', {'dht-enabled': bool(value)})


    @_setting(Bool)
    def pex(self):
        """Whether to use PEX to discover peers for public torrents"""
        return self._get('pex', 'pex-enabled')

    async def get_pex(self):
        """Refresh cache and return `pex`"""
        await self.update()
        return self.pex

    async def set_pex(self, enabled):
        """See `pex`"""
        value = self._converters['pex'](enabled)
        await self._set('pex', {'pex-enabled': bool(value)})


    @_setting(Bool)
    def lpd(self):
        """Whether to use LPD to discover peers for public torrents"""
        return self._get('lpd', 'lpd-enabled')

    async def get_lpd(self):
        """Refresh cache and return `lpd`"""
        await self.update()
        return self.lpd

    async def set_lpd(self, enabled):
        """See `lpd`"""
        value = self._converters['lpd'](enabled)
        await self._set('lpd', {'lpd-enabled': bool(value)})


    # Local Filesystem settings

    @_setting(Path)  # noqa: E301
    def path_complete(self):
        """Where to put downloaded files"""
        return self._get('path.complete', 'download-dir')

    async def get_path_complete(self):
        """Refresh cache and return `path_complete`"""
        await self.update()
        return self.path_complete

    async def set_path_complete(self, path):
        """
        Set download directory files

        If path is relative (i.e. doesn't start with a path separator
        (e.g. '/')), it is relative to what `get_path_complete` returns.
        """
        value = self._converters['path.complete'](path)
        if not value.startswith(os.sep):
            current_path = await self.get_path_complete()
            value = os.path.join(current_path, value)
        await self._set('path.complete', {'download-dir': value})


    @_setting(BoolOrPath,
              description='Where to put partially downloaded files')
    def path_incomplete(self):
        """
        Path to incomplete files or Bool(<False>) to put them in `path_complete`
        """
        def get_raw_value():
            if self._raw['incomplete-dir-enabled']:
                return self._raw['incomplete-dir']
            else:
                return False
        return self._get('path.incomplete', get_raw_value)

    async def get_path_incomplete(self):
        """Refresh cache and return `path_incomplete`"""
        await self.update()
        return self.path_incomplete

    async def set_path_incomplete(self, path):
        """
        Set path to incomplete files or Bool(<False>)

        If `path` is not a `str` instance, it is evaluated as a bool and this
        feature is enabled or disabled accordingly without changing the path.

        If path is relative (i.e. doesn't start with a path separator
        (e.g. '/')), it is relative to what `get_path_incomplete` returns.
        """
        value = self._converters['path.incomplete'](path)
        request = {'incomplete-dir-enabled': bool(value)}
        if isinstance(value, Path):
            if not value.startswith(os.sep):
                current_path = await self.get_path_incomplete()
                value = os.path.join(current_path, value)
            request['incomplete-dir'] = str(value)
        await self._set('path.incomplete', request)


    @_setting(Bool)
    def files_part(self):
        """Whether ".part" is appended to incomplete files"""
        return self._get('files.part', 'rename-partial-files')

    async def get_files_part(self):
        """Refresh cache and return `files_part`"""
        await self.update()
        return self.files_part

    async def set_files_part(self, enabled):
        """See `files_part`"""
        value = self._converters['files.part'](enabled)
        await self._set('files.part', {'rename-partial-files': bool(value)})


    # Rate limits

    @staticmethod  # noqa: E301
    def _limit_rate_fields(direction, alt):
        if alt:
            field_value = 'alt-speed-' + direction
            field_enabled = 'alt-speed-enabled'
        else:
            field_value = 'speed-limit-' + direction
            field_enabled = 'speed-limit-' + direction + '-enabled'
        return (field_value, field_enabled)

    @staticmethod
    def _limit_rate_key(direction, alt):
        if alt:
            return 'limit.rate.alt.%s' % direction
        else:
            return 'limit.rate.%s' % direction

    def _get_limit_rate(self, direction, alt=False):
        key = self._limit_rate_key(direction, alt)
        if self._cache[key] is None:
            field_value, field_enabled = self._limit_rate_fields(direction, alt)
            if self._raw is None:
                self._cache[key] = const.DISCONNECTED
            elif self._raw[field_enabled]:
                # Transmission reports kilobytes
                self._cache[key] = self._converters[key](
                    convert.bandwidth(self._raw[field_value] * 1000, unit='B'))
            else:
                self._cache[key] = const.UNLIMITED
        return self._cache[key]

    async def _set_limit_rate(self, limit, direction, alt=False):
        key = self._limit_rate_key(direction, alt)
        field_value, field_enabled = self._limit_rate_fields(direction, alt)
        limit = self._converters[key](limit)
        if isinstance(limit, Bool):
            await self._set(key, {field_enabled: bool(limit)})
        elif 0 <= limit < float('inf'):
            raw_limit = round(round(limit.copy(convert_to='B')) / 1000)
            await self._set(key, {field_enabled: True,
                                  field_value: raw_limit})
        else:
            await self._set(key, {field_enabled: False})

    async def _adjust_limit_rate(self, adjustment, direction, alt=False):
        key = self._limit_rate_key(direction, alt)
        field_value, field_enabled = self._limit_rate_fields(direction, alt)
        prop_name = _key2property(key)

        current_limit = getattr(self, prop_name)
        if current_limit is const.DISCONNECTED:
            current_limit = await getattr(self, 'get_' + prop_name)()
        adjustment = self._converters[key](adjustment)
        new_limit = BoolOrBandwidth.adjust(current_limit, adjustment)
        await self._set_limit_rate(new_limit, direction, alt=alt)

    @_setting(BoolOrBandwidth,
              description='Global upload rate limit')
    def limit_rate_up(self):
        """
        Upload rate limit

        This uses the application-wide `stig.utils.convert` module to get
        consistent units and unit prefixes.
        """
        return self._get_limit_rate('up', alt=False)

    async def get_limit_rate_up(self):
        """Refresh cache and return `limit_rate_up`"""
        await self.update()
        return self.limit_rate_up

    async def set_limit_rate_up(self, limit):
        """
        Set upload rate limit to `limit` (see also `limit_rate_up`)

        An existing limit is disabled if `limit` is False, the constant
        UNLIMITED, or a number < 0.

        A previously set and then disabled limit is enabled if `limit` is True.
        """
        await self._set_limit_rate(limit, 'up', alt=False)

    async def adjust_limit_rate_up(self, limit):
        """Adjust current upload rate limit by `limit` (positive or negative number)"""
        await self._adjust_limit_rate(limit, 'up', alt=False)


    @_setting(BoolOrBandwidth,
              description='Global download rate limit')
    def limit_rate_down(self):
        """Download rate limit (see `limit_rate_up`)"""
        return self._get_limit_rate('down', alt=False)

    async def get_limit_rate_down(self):
        """Refresh cache and return `limit_rate_down`"""
        await self.update()
        return self.limit_rate_down

    async def set_limit_rate_down(self, limit):
        """Set download rate limit to `limit` (see `set_limit_rate_up`)"""
        await self._set_limit_rate(limit, 'down', alt=False)

    async def adjust_limit_rate_down(self, limit):
        """Adjust current download rate limit by `limit` (positive or negative number)"""
        await self._adjust_limit_rate(limit, 'down', alt=False)


    @_setting(BoolOrBandwidth,
              description='Alternative global upload rate limit')
    def limit_rate_alt_up(self):
        """Alternative upload rate limit (see `limit_rate_up`)"""
        return self._get_limit_rate('up', alt=True)

    async def get_limit_rate_alt_up(self):
        """Refresh cache and return `limit_rate_alt_up`"""
        await self.update()
        return self.limit_rate_alt_up

    async def set_limit_rate_alt_up(self, limit):
        """Set alternative upload rate limit to `limit` (see `set_limit_rate_up`)"""
        await self._set_limit_rate(limit, 'up', alt=True)

    async def adjust_limit_rate_alt_up(self, limit):
        """Adjust current alternative upload rate limit by `limit`"""
        await self._adjust_limit_rate(limit, 'up', alt=True)


    @_setting(BoolOrBandwidth,
              description='Alternative global download rate limit')
    def limit_rate_alt_down(self):
        """Alternative download rate limit (see `limit_rate_up`)"""
        return self._get_limit_rate('down', alt=True)

    async def get_limit_rate_alt_down(self):
        """Refresh cache and return `limit_rate_alt_down`"""
        await self.update()
        return self.limit_rate_alt_down

    async def set_limit_rate_alt_down(self, limit):
        """Set alternative upload rate limit to `limit` (see `set_limit_rate_up`)"""
        await self._set_limit_rate(limit, 'down', alt=True)

    async def adjust_limit_rate_alt_down(self, limit):
        """Adjust current alternative download rate limit by `limit`"""
        await self._adjust_limit_rate(limit, 'down', alt=True)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import re
from collections import abc
from typing import Pattern

from natsort import humansorted


class Categories(abc.Sequence):
    """Iterable over non-empty Candidates objects with selection tracking"""

    def __init__(self, *categories):
        self._categories = categories
        self._current_index = 0 if categories else None

    def __getitem__(self, key):
        return tuple(self)[key]

    def __iter__(self):
        for cat in self._categories:
            if len(cat) > 0:
                yield cat

    def __len__(self):
        return len(tuple(iter(self)))

    def next(self):
        """
        Select next candidate in current Candidates or first candidate in next
        Candidates or first candidate in first Candidates
        """
        cats = tuple(self)
        if not cats:
            self.current_index = None
        else:
            if self.current and self.current.current_index < len(self.current) - 1:
                # Move to next candidate in current list
                self.current.next()
            elif self.current_index < len(cats) - 1:
                # Move to first candidate in next list
                self.current_index += 1
                self.current.current_index = 0
            else:
                # Wrap to first candidate in first list
                self.current_index = self.current.current_index = 0

    def prev(self):
        """
        Select previous candidate in current Candidates or last candidate in
        previous Candidates or last candidate in last Candidates
        """
        cats = tuple(self)
        if not cats:
            self.current_index = None
        else:
            if self.current and self.current.current_index > 0:
                # Move to previous candidate in current list
                self.current.prev()
            elif self.current_index > 0:
                # Move to last candidate in previous list
                self.current_index -= 1
                self.current.current_index = len(self.current) - 1
            else:
                # Wrap to last candidate in last list
                self.current_index = len(cats)
                self.current.current_index = len(self.current) - 1

    @property
    def all(self):
        """All Candidates, including empty ones"""
        return self._categories

    @property
    def current_index(self):
        """Index of currently selected Candidates object or None"""
        # Make sure the current index is not pointing to an empty list
        self.current_index = self._current_index
        return self._current_index

    @current_index.setter
    def current_index(self, index):
        if not self:
            self._current_index = None
        elif index is None and self:
            self._current_index = 0
        else:
            self._current_index = max(0, min(index, len(self) - 1))

    @property
    def current(self):
        """Currently selected Candidates object or None"""
        i = self.current_index
        if i is not None:
            return tuple(self)[i]

    def __repr__(self):
        return '%s(%s)' % (type(self).__name__,
                           ', '.join(repr(cands) for cands in self._categories))


class Candidates(abc.Sequence):
    """Sequence of completion candidates"""

    def __init__(self, candidates=(), label='', curarg_seps=()):
        # Remove duplicates while preserving order:
        # https://www.peterbe.com/plog/fastest-way-to-uniquify-a-list-in-python-3.6
        cands_typed = (cand if isinstance(cand, Candidate) else Candidate(cand)
                       for cand in candidates)
        cands_deduped = (cand for cand in dict.fromkeys(cands_typed))
        cands_noempty = (cand for cand in cands_deduped if cand != '')
        cands_sorted = humansorted(cands_noempty)
        self._candidates = tuple(cands_sorted)
        self._matches = self._candidates
        self._curarg_seps = tuple(sorted(curarg_seps))
        self._label = str(label)
        self._current_index = 0 if self._candidates else None

    def __getitem__(self, key):
        return self._matches[key]

    def __len__(self):
        return len(self._matches)

    def next(self):
        """Select next candidate or first one if last is selected"""
        if not self:
            self._current_index = None
        elif self._current_index >= len(self) - 1:
            self._current_index = 0
        else:
            self._current_index += 1

    def prev(self):
        """Select previous candidate or last one if last is selected"""
        if not self:
            self._current_index = None
        elif self._current_index <= 0:
            self._current_index = len(self) - 1
        else:
            self._current_index -= 1

    @property
    def current_index(self):
        """Index of currently selected candidate or None"""
        return self._current_index

    @current_index.setter
    def current_index(self, index):
        if not self._matches:
            self._current_index = None
        elif index is None and self._matches:
            self._current_index = 0
        else:
            self._current_index = max(0, min(index, len(self._matches) - 1))

    @property
    def current(self):
        """Currently selected candidate or None"""
        if self._current_index is not None:
            return self[self._current_index]

    def reduce(self, regex):
        """Reduce the candidates to the ones that match `pattern`"""
        if not isinstance(regex, Pattern):
            regex = re.compile(regex)
        # Remember current candidate so we can keep it selected if possible
        curcand = self.current
        self._matches = tuple(cand for cand in self._candidates
                              if re.search(regex, cand))
        if curcand in self._matches:
            self.current_index = self._matches.index(curcand)
        elif self._matches:
            self.current_index = 0
        else:
            self.current_index = None

    @property
    def curarg_seps(self):
        """
        List of strings at which the current argument should be split before
        applying a candidate from this list
        """
        return self._curarg_seps

    @curarg_seps.setter
    def curarg_seps(self, seps):
        self._curarg_seps = seps

    @property
    def label(self):
        """Category or short description"""
        return self._label

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self._candidates == other._candidates and
                self._label == other._label and
                self._curarg_seps == other._curarg_seps)

    def __hash__(self):
        return hash(self._candidates)

    def __repr__(self):
        kwargs = {}
        if self.curarg_seps: kwargs['curarg_seps'] = self.curarg_seps
        if self.label: kwargs['label'] = self.label
        if kwargs:
            return '%s(%r, %s)' % (
                type(self).__name__, self._candidates,
                ', '.join('%s=%r' % (k,v) for k,v in kwargs.items())
            )
        else:
            return '%s(%r)' % (type(self).__name__, self._candidates)


class Candidate(str):
    """A string with some attributes"""

    def __new__(cls, string, **kwargs):
        return super().__new__(cls, string)

    def __init__(self, string, in_parens='', **info):
        # `in_parens` is displayed in parentheses after `string`.
        # `info` is a mapping with extra information, e.g. `Description='This
        # value does that'` or 'Default='<default value>'`.
        self.info = {}
        for title,text in info.items():
            self.info[str(title)] = str(text)
        self.in_parens = str(in_parens)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return super().__eq__(other)

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        kwargs = {}
        if self.in_parens: kwargs['in_parens'] = self.in_parens
        for title,text in self.info.items():
            kwargs[title] = text
        if kwargs:
            return '%s(%r, %s)' % (
                type(self).__name__, str(self),
                ', '.join('%s=%r' % (k,v) for k,v in kwargs.items())
            )
        else:
            return '%s(%r)' % (type(self).__name__, str(self))


class SingleCandidate(Candidates):
    """
    Dummy Candidates that contains only one replaceable candidate

    This is used to include current user input in its own category.
    """

    def __init__(self, string, curarg_seps=()):
        super().__init__((string,), curarg_seps=curarg_seps)
        if not self._candidates:
            self._candidates = self._matches = (Candidate(''),)
            self._current_index = 0

    def set(self, string):
        self._candidates = self._matches = \
            (Candidate(string) if not isinstance(string, Candidate) else string,)

    def reduce(self, *args, **kwargs):
        pass

    def __hash__(self):
        return hash(id(self))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import builtins
from collections import abc
from functools import lru_cache

import urwid
import urwidtrees
from natsort import humansorted
from urwidtrees.decoration import ArrowTree

from ...client import FileFilter
from ...views.file import TorrentFileDirectory
from .base import ItemWidgetBase, ListWidgetBase
from .file import TUICOLUMNS
from .utils import stringify_torrent_filter

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


class FileTreeDecorator(ArrowTree):
    """urwidtrees decorator for TorrentFiles and TorrentFileTrees"""

    def __init__(self, torrents, keymap, table, ffilter):
        self._filewidgetcls = keymap.wrap(FileItemWidget, context='file')
        self._table = table
        self._ffilter = ffilter
        self._widgets = {}
        self._filtered_counts = {}
        forest = self._create_file_forest(torrents)
        super().__init__(forest, indent=2)

    def _file_is_filtered(self, tfile):
        if self._ffilter is None:
            return False  # No filter specified
        elif isinstance(self._ffilter, (abc.Sequence, abc.Set)):
            # ffilter is a collection of file IDs
            return not tfile['id'] in self._ffilter
        else:
            # ffilter is a FileFilter instance
            return not self._ffilter.match(tfile)

    def _create_file_forest(self, torrents):
        # Create a list of nested trees in SimpleTree format.  But the leaves
        # are mappings instead of widgets.  Each mapping contains the
        # information that the `decorate` method needs to create a widget.

        def create_tree(node, content):
            if content.nodetype == 'leaf':
                # Process single file
                if not self._file_is_filtered(content):
                    return (content, None)
                else:
                    return None

            elif content.nodetype == 'parent':
                # Recurse into directory entries (files and subdirectories)
                tree = []
                filtered_count = 0
                content_sorted = humansorted(content.items())
                for k,v in content_sorted:
                    if v.nodetype == 'leaf':
                        if not self._file_is_filtered(v):
                            tree.append((v, None))
                            nonlocal fcount
                            fcount += 1
                        else:
                            filtered_count += 1
                    elif v.nodetype == 'parent':
                        dirnode = TorrentFileDirectory(name=k, tree=v)
                        tree.append(create_tree(dirnode, v))

                self._filtered_counts[node['id']] = filtered_count
                node['name'] = TorrentFileDirectory.create_directory_name(node['name'], filtered_count)
                return (node, tree or None)

        fcount = 0
        forest = []  # Multiple trees as siblings
        for t in humansorted(torrents, key=lambda t: t['name']):
            filetree = t['files']
            filecount = len(tuple(filetree.files))
            if filecount > 0:
                nodetree = None
                rootnodename = next(iter(filetree.keys()))
                if len(tuple(filetree.directories)) == 0:
                    # Single-file torrent
                    if not self._file_is_filtered(filetree[rootnodename]):
                        nodetree = (filetree[rootnodename], None)
                else:
                    # Torrent with directory structure
                    # If we don't ignore the topmost node, the torrent itself is
                    # displayed as a directory.
                    content = filetree[rootnodename]
                    rootnode = TorrentFileDirectory(rootnodename, tree=content)
                    nodetree = create_tree(rootnode, content)

                if nodetree is not None:
                    forest.append(nodetree)

        self.filecount = fcount
        return forest

    @lru_cache(maxsize=1024)
    def decorate(self, pos, data, is_first=True):
        # We can use the tree position as table ID
        self._table.register(pos)
        row = self._table.get_row(pos)

        # We use parent's decorate() method to give the name column a tree
        # structure.  But we also need the original update() method so we can
        # apply new data to the widget.  This is dirty but it works.
        if row.exists('name'):
            update_method = row.name.update
            decowidget = super().decorate(pos, row.name, is_first=is_first)
            decowidget.update = update_method
            row.replace('name', decowidget)

        # Wrap the whole row in a FileItemWidget with keymapping.  This also
        # applies all the other values besides the name (size, progress, etc).
        file_widget = self._filewidgetcls(data, row)
        self._widgets[data['id']] = file_widget
        return file_widget

    def update(self, torrents):
        widgets = self._widgets
        for t in torrents:
            # Update file nodes
            for f in t['files'].files:
                node_id = f['id']
                # File might be filtered
                if node_id in widgets:
                    widgets[node_id].update(f)

            # Update directory nodes
            for name,content in t['files'].directories:
                node_id = content.id
                if node_id in widgets:
                    filtered_count = self._filtered_counts[node_id]
                    data = TorrentFileDirectory(name, tree=content,
                                                filtered_count=filtered_count)
                    widgets[node_id].update(data)

    @property
    def widgets(self):
        """Yield all file and directory widgets in this tree"""
        yield from self._widgets.values()


class FileItemWidget(ItemWidgetBase):
    palette_unfocused = 'filelist'
    palette_focused   = 'filelist.focused'
    columns_focus_map = {}
    for col in TUICOLUMNS.values():
        columns_focus_map.update(col.style.focus_map)

    @property
    def id(self):
        return self.data['id']

    @property
    def torrent_id(self):
        return self.data['tid']

    @property
    def nodetype(self):
        """'parent' or 'leaf'"""
        return self.data.nodetype


class FileListWidget(ListWidgetBase):
    tuicolumns      = TUICOLUMNS
    ListItemClass   = FileItemWidget
    keymap_context  = 'file'
    palette_name    = 'filelist'
    focusable_items = True

    def __init__(self, srvapi, keymap, tfilter, ffilter, columns=None, title=None):
        super().__init__(srvapi, keymap, columns=columns, title=title)
        self._tfilter = tfilter
        self._ffilter = ffilter
        self._secondary_filter = None
        self._initialized = False
        self._torrents = None

        self._poller = self._srvapi.create_poller(
            self._srvapi.torrent.torrents, tfilter, keys=('files', 'name')
        )
        self._poller.on_response(self._handle_files)

    def _handle_files(self, response):
        if response is None or not response.torrents:
            self.clear()
        else:
            if self._initialized:
                self._update_listitems(response.torrents)
            else:
                self._init_listitems(response.torrents)
                self._initialized = True
        self._invalidate()

    def _init_listitems(self, torrents):
        self.clear()
        if torrents:
            self._torrents = torrents
            self._create_filetree()
            self._listbox._invalidate()

    def _create_filetree(self):
        # Combine primary and secondary file filters
        ffilter = self._ffilter
        sffilter = self._secondary_filter
        if ffilter is None:
            ffilter = sffilter
        elif sffilter is not None:
            ffilter = ffilter & sffilter

        self._filetree = FileTreeDecorator(self._torrents, self._keymap, self._table, ffilter)
        self._listbox.body = urwidtrees.widgets.TreeListWalker(self._filetree)

    def _update_listitems(self, torrents=()):
        if torrents:
            self._filetree.update(torrents)
            self._torrents = torrents

    @property
    def secondary_filter(self):
        return self._secondary_filter

    @secondary_filter.setter
    def secondary_filter(self, file_filter):
        if file_filter is None:
            self._secondary_filter = None
        else:
            self._secondary_filter = FileFilter(file_filter)
        self._create_filetree()


    @property
    def title(self):
        if self._title_name is None:
            if self._torrents is None:
                return ''
            else:
                title = stringify_torrent_filter(self._tfilter, self._torrents)
                if self._ffilter:
                    title += ' %s' % self._ffilter
                return title
        else:
            return super().title

    def clear(self):
        # We can't call super().clear() because it runs the more efficient
        # `self._listbox.body[:] = ()`
        # That doesn't work here because urwidtrees.TreeListWalker doesn't
        # support item assignment.
        self._listbox.body = urwid.SimpleListWalker([])
        self._listbox._invalidate()
        self._initialized = False
        self._table.clear()
        self._marked.clear()

    def refresh(self):
        self._poller.poll()

    @property
    def count(self):
        return self._filetree.filecount if hasattr(self, '_filetree') else 0

    @property
    def focus_position(self):
        positions = tuple(self._filetree.positions())
        return positions.index(self._listbox.focus_position)

    @focus_position.setter
    def focus_position(self, focus_position):
        positions = tuple(self._filetree.positions())
        i = min(focus_position, len(positions) - 1)
        try:
            self._listbox.focus_position = positions[i]
        except KeyError:
            pass

    @property
    def focused_file_ids(self):
        """File IDs of the focused files in a tuple"""
        focused = self.focused_widget
        if focused is not None:
            # The focused widget in the list can be a file or a directory.  If
            # it's a directory, the 'id' property returns the IDs of all the
            # contained files recursively, otherwise it's just a single ID.  For
            # consistency, we always return a sequence of IDs.
            if focused.nodetype == 'leaf':
                return (focused.id,)
            else:
                return focused.id

    @property
    def focused_torrent_id(self):
        """Torrent ID of the currently focused file or `None`"""
        focused_widget = self.focused_widget
        if focused_widget is not None:
            return focused_widget.torrent_id


    def all_children(self, pos):
        """Yield (position, widget) tuples of all sub-nodes (leaves and parents)"""
        ft = self._filetree
        lb = self._listbox

        def recurse(subpos):
            widget = lb.body[subpos]
            if ft.is_leaf(subpos):
                yield (subpos, widget)
            else:
                # Yield sub-parent nodes, but not the starting node that was
                # passed to us
                if subpos != pos:
                    yield (subpos, widget)

                new_subpos = ft.first_child_position(subpos)
                while new_subpos is not None:
                    yield from recurse(new_subpos)
                    new_subpos = ft.next_sibling_position(new_subpos)

        yield from recurse(pos)

    def _set_mark(self, mark, toggle=False, all=False):
        if toggle:
            focused = self.focused_widget
            if focused is not None:
                mark = not focused.is_marked

        def get_widget(pos):
            return self._listbox.body[pos]

        def mark_leaves(pos, mark):
            get_widget(pos).is_marked = mark

            for subpos,widget in self.all_children(pos):
                if widget.nodetype == 'leaf':
                    widget.is_marked = mark
                    if mark:
                        self._marked.add(widget)
                    else:
                        self._marked.discard(widget)

                elif widget.nodetype == 'parent':
                    if pos != subpos:  # Avoid infinite recursion
                        mark_leaves(subpos, mark)

        if all:
            # Top ancestor node positions are (0,), (1,), (3,) etc
            for pos in self._filetree.positions():
                if len(pos) == 1:
                    mark_leaves(pos, mark)
        else:
            mark_leaves(self._listbox.focus_position, mark)
        assert builtins.all(m.nodetype == 'leaf' for m in self._marked)

        # A parent node is marked only if all its children are marked.  To check
        # that, we walk through every ancestor up to the top and check all its
        # children.  There is no need to check the children of other parent
        # nodes (uncles, great uncles, etc) because they should already be
        # marked properly from previous runs.

        def all_children_marked(pos):
            marked = True
            childpos = self._filetree.first_child_position(pos)
            while childpos is not None:
                marked = marked and get_widget(childpos).is_marked
                childpos = self._filetree.next_sibling_position(childpos)
            return marked

        parpos = self._filetree.parent_position(self._listbox.focus_position)
        while parpos is not None:
            parwidget = get_widget(parpos)
            parwidget.is_marked = all_children_marked(parpos)
            parpos = self._filetree.parent_position(parpos)

    def refresh_marks(self):
        for widget in self._filetree.widgets:
            widget.is_marked = widget.is_marked



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import asyncio
import calendar
import datetime
import operator
import os
import re
import time
from types import SimpleNamespace

from async_timeout import timeout as async_timeout

from ..utils import cached_property, convert  # noqa: F401
from ..utils.usertypes import (Bool, Float, Int, Option, Path, Percent,  # noqa: F401
                               String, multitype)
from . import constants as const


class SHA1(String):
    def __new__(cls, value, regex=r'^[0-9a-fA-F]{,40}$'):
        return super().__new__(cls, value, regex=regex)


class SizeInBytes(Int):
    def __new__(cls, value, **kwargs):
        return convert.size(value, unit='byte')


class Bandwidth(Int):
    typename = 'bandwidth'

    def __new__(cls, value, **kwargs):
        value = convert.bandwidth(value)
        kwargs.update(unit=convert.bandwidth.unit,
                      prefix=convert.bandwidth.prefix)
        return super().__new__(cls, value, **kwargs)

    @classmethod
    def _get_syntax(cls, **kwargs):
        return '%s[b|B]' % super()._get_syntax(**kwargs)


class BandwidthInBytes(Bandwidth):
    def __new__(cls, value, **kwargs):
        return convert.bandwidth(value, unit='byte')


class BandwidthInBytesOrNone(BandwidthInBytes):
    def __new__(cls, value, **kwargs):
        if value is None:
            return const.UNLIMITED
        else:
            return super().__new__(cls, value, **kwargs)


class BoolOrBandwidth(multitype(Bool.partial(true=('limited', 'enabled', 'yes', 'on', 'true'),
                                             false=('unlimited', 'disabled', 'no', 'off', 'false')),
                                Bandwidth)):
    @staticmethod
    def adjust(current, adjustment):
        """Adjust `current` by `adjustment`"""
        if isinstance(current, Bool) or current >= float('inf'):
            # If current number is infinity, adjust from 0
            current = 0
        new = current + adjustment
        if new < 0:
            # Drop to 0 if current is greater than zero.
            # If current already is zero, drop to infinity.
            if current > 0:
                new = 0
            else:
                new = const.UNLIMITED
        return new

    def __new__(cls, value, **kwargs):
        if isinstance(value, (float, int)) and value >= float('inf'):
            return super().__new__(cls, 'unlimited')
        else:
            return super().__new__(cls, value, **kwargs)


class SmartCmpStr(str):
    """
    String that compares case-insensitively if the other string consists solely
    of lower case characters.
    """
    def __cmp(self, op, other):
        if not isinstance(other, str):
            return NotImplemented
        o = str(other)
        if o == o.casefold():
            s = self.casefold()
        else:
            s = str(self)
        return op(s, o)

    def __lt__(self, other): return self.__cmp(operator.lt, other)
    def __le__(self, other): return self.__cmp(operator.le, other)
    def __eq__(self, other): return self.__cmp(operator.eq, other)
    def __ne__(self, other): return self.__cmp(operator.ne, other)
    def __gt__(self, other): return self.__cmp(operator.gt, other)
    def __ge__(self, other): return self.__cmp(operator.ge, other)
    def __contains__(self, other): return self.__cmp(operator.contains, other)

    # Defining __eq__ mandates defining __hash__ to make instances hashable
    def __hash__(self):
        return super().__hash__()


class SmartCmpPath(SmartCmpStr):
    def __new__(cls, path):
        return super().__new__(cls, os.path.normpath(path))


BoolOrPath = multitype(Bool, Path)


class Count(Int):
    UNKNOWN = -1

    def __str__(self):
        return '?' if self == self.UNKNOWN else super().without_unit


class Ratio(Float):
    """A Torrent's upload/download ratio as a float"""
    INFINITE = float('inf')
    NOT_APPLICABLE = -1

    def __str__(self):
        if self == self.INFINITE:
            return '∞'
        elif self == self.NOT_APPLICABLE:
            return ''
        else:
            return super().without_unit


class Status(tuple):
    """A Torrent's status as a tuple of strings"""

    IDLE      = 'idle'
    DOWNLOAD  = 'downloading'
    UPLOAD    = 'uploading'
    CONNECTED = 'connected'
    SEED      = 'seeding'
    STOPPED   = 'stopped'
    QUEUED    = 'queued'
    ISOLATED  = 'isolated'
    VERIFY    = 'verifying'
    INIT      = 'discovering'
    ORDER = (VERIFY, DOWNLOAD, UPLOAD, INIT, CONNECTED,
             ISOLATED, QUEUED, IDLE, STOPPED, SEED)

    def __lt__(self, other):
        return self.ORDER.index(self[0]) < self.ORDER.index(other[0])

    def __le__(self, other):
        return self.ORDER.index(self[0]) <= self.ORDER.index(other[0])

    def __gt__(self, other):
        return self.ORDER.index(self[0]) > self.ORDER.index(other[0])

    def __ge__(self, other):
        return self.ORDER.index(self[0]) >= self.ORDER.index(other[0])


SECONDS = (('y', 31557600),  # 365.25 days
           ('M',  2629800),  # 1 year / 12
           ('w',   604800),  # 7 days
           ('d',    86400),
           ('h',     3600),
           ('m',       60),
           ('s',        1))

class Timedelta(int):
    # To sort unknown and not applicable Timedeltas below the rest, these
    # constants have large values that are very likely never encountered as
    # actual values.
    UNKNOWN        = 1e10
    NOT_APPLICABLE = 1e11
    CONSTANTS = (UNKNOWN, NOT_APPLICABLE)
    _CONSTANTS_MAP_STRINGS = {'unknown': UNKNOWN, 'na': NOT_APPLICABLE,
                              'n/a': NOT_APPLICABLE, 'not applicable': NOT_APPLICABLE}

    _FULL_REGEX = re.compile((r'^(in |-|\+|)(\S+)( ago|)$'), flags=re.IGNORECASE)
    _SPLIT_REGEX = re.compile((r'((?:\d+\.\d+|\d+|\.\d+)[' +
                               r''.join(unit for unit,secs in SECONDS) +
                               r']?)'))

    @classmethod
    def from_string(cls, string):
        const_value = cls._CONSTANTS_MAP_STRINGS.get(string, None)
        if const_value is not None:
            return cls(const_value)

        match = cls._FULL_REGEX.match(string)
        if not match:
            raise ValueError('Invalid %s: %r' % (cls.__name__, string))

        sign_start = match.group(1).lower().strip()
        timespan = match.group(2)
        sign_end = match.group(3).lower().strip()
        sign = 1  # 1/-1 to represent +/-

        if sign_start == '-' or sign_end == 'ago':
            if sign_start == '+':
                raise ValueError('Invalid %s: %r' % (cls.__name__, string))
            sign = -1
        if any(sign_start == x for x in ('+', 'in')):
            if sign_end == 'ago':
                raise ValueError('Invalid %s: %r' % (cls.__name__, string))

        secs_total = 0
        kwargs = {}
        for part in cls._SPLIT_REGEX.split(timespan):
            if len(part) < 1:
                continue
            elif not cls._SPLIT_REGEX.match(part):
                raise ValueError('Invalid %s: %r' % (cls.__name__, string))
            elif part[-1].isdigit():
                # No unit specified - assume seconds
                secs_total += float(part)
            else:
                num, unit = part[:-1], part[-1]
                for unit_,secs in SECONDS:
                    if unit == unit_:
                        secs_total += float(num) * secs
                        if unit == 'y':
                            kwargs['_real_years'] = float(num) * sign
                        elif unit == 'M':
                            kwargs['_real_months'] = float(num) * sign
                        break

        return cls(secs_total * sign, **kwargs)

    def __new__(cls, seconds, _real_years=None, _real_months=None):
        obj = super().__new__(cls, seconds)
        obj._real_years = _real_years
        obj._real_months = _real_months
        return obj

    @cached_property
    def inverse(self):
        """Return the same object with the sign switched"""
        real_years, real_months = self._real_years, self._real_months
        return type(self)(-int(self),
                          _real_years=-real_years if real_years is not None else None,
                          _real_months=-real_months if real_months is not None else None)

    def __str__(self):
        if self == self.UNKNOWN:
            return '?'
        elif self == self.NOT_APPLICABLE:
            return ''
        elif self == 0:
            return '0s'

        abs_secs = abs(self)
        for i,(unit,amount) in enumerate(SECONDS):
            if abs_secs >= amount:
                num = self / amount

                # Small numbers get a sub-unit, for example '1d15h'
                if 1 <= abs_secs / amount < 10 and i < len(SECONDS) - 1:
                    subunit, subamount = SECONDS[i + 1]
                    if num >= 0:
                        subnum = abs(((num % 1) * amount) / subamount)
                    else:
                        subnum = abs(((num % -1) * amount) / subamount)

                    if subnum >= 1:
                        return '%d%s%d%s' % (int(num), unit, int(subnum), subunit)

                return '%d%s' % (int(num), unit)

    @cached_property
    def with_preposition(self):
        if self > 0:
            return 'in %s' % self
        elif self < 0:
            return ('%s ago' % self)[1:]  # Remove the first char ('-')
        else:
            return 'now'

    def __bool__(self):
        """Whether delta is known"""
        return self not in self.CONSTANTS

    @cached_property
    def is_known(self):
        return bool(self)

    @property
    def timestamp(self):
        if self == self.UNKNOWN:
            return Timestamp(Timestamp.UNKNOWN)
        elif self == self.NOT_APPLICABLE:
            return Timestamp(Timestamp.NOT_APPLICABLE)
        else:
            ts = None
            real_years, real_months = self._real_years, self._real_months
            if real_years or real_months:
                # Because we can't know the exact number of seconds in a
                # year/month, we take the number of years/months provided by the
                # parser and use datetime and timedelta to calculate the
                # timestamp.
                secs = int(self)
                dt = datetime.datetime.now()
                # datetime will throw OverflowError if numbers are too large
                try:
                    if real_years is not None:
                        dt = dt.replace(year=int(dt.year + real_years))
                        # Subtract the approximate number seconds in the number of years
                        secs = secs % (real_years * SECONDS[0][1])

                    if real_months is not None:
                        # When adding/subtracting months, we may have to keep the
                        # month between 1-12. We also may have to adjust the year
                        # and/or day.
                        # https://stackoverflow.com/a/4131114
                        month = int(dt.month - 1 + real_months)
                        year = int(dt.year + month // 12)  # Integer division
                        month = int(month % 12 + 1)
                        day = min(dt.day, calendar.monthrange(year, month)[1])  # Avoid day 30 in February
                        dt = dt.replace(year=year, month=month, day=day)
                        # Subtract the approximate number seconds in the number of months
                        secs = secs % (real_months * SECONDS[1][1])
                except (OverflowError, ValueError):
                    pass
                else:
                    dt = dt + datetime.timedelta(seconds=secs)
                    ts = Timestamp(dt.timestamp())

            if ts is None:
                # Without years or months given, we can simply add our delta to time.time()
                ts = Timestamp(time.time() + self)
            return ts

    def __repr__(self):
        return '<%s %s / %s>' % (type(self).__name__, str(int(self)), self.__str__())


class Timestamp(float):
    # These constants get "random" fractions added to make it less likely that
    # any real-world value equals them.
    NOW            = -2 + 0.123456789
    SOON           = -1 + 0.123456789
    UNKNOWN        = 1e10 + 0.123456789
    NOT_APPLICABLE = 1e11 + 0.123456789
    NEVER          = -1e12 + 0.123456789
    CONSTANTS = (NOW, SOON, UNKNOWN, NOT_APPLICABLE, NEVER)
    _CONSTANTS_MAP_STRINGS = {'now': NOW, 'soon': SOON, 'unknown': UNKNOWN,
                              'na': NOT_APPLICABLE, 'n/a': NOT_APPLICABLE, 'not applicable': NOT_APPLICABLE,
                              'never': NEVER}

    _FORMATS_DATE = (('%Y',       ('year',)),
                     ('%Y-%m',    ('year', 'month')),
                     ('%Y-%m-%d', ('year', 'month', 'day')),
                     ('%d',       ('day',)),
                     ('%m-%d',    ('month', 'day')))
    _FORMATS_TIME = (('%H:%M',    ('hour', 'minute')),
                     ('%H:%M:%S', ('hour', 'minute', 'second')))

    # Create all combinations of date, time and date+time formats, keeping track
    # of the values they specify
    _FORMATS = []
    for date_frmt,date_given in _FORMATS_DATE:
        _FORMATS.append((date_frmt, date_given))
        for time_frmt,time_given in _FORMATS_TIME:
            _FORMATS.append((time_frmt, time_given))
            given = date_given + time_given
            _FORMATS.append(('%s %s' % (date_frmt, time_frmt), given))
            _FORMATS.append(('%s %s' % (time_frmt, date_frmt), given))

    @classmethod
    def from_string(cls, string):
        string = string.strip().replace('  ', ' ')
        now = datetime.datetime.now()
        value = cls._CONSTANTS_MAP_STRINGS.get(string.lower(), None)
        if value is not None:
            return cls(value, _timerange=(value, value))

        dt_now = now.replace(second=0, microsecond=0)
        dt_default = now.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)

        def fill_in_missing_values(dt, given):
            names = ('year', 'month', 'day', 'hour', 'minute', 'second')
            most_significant_given = names.index(given[0])
            new_args = {}
            for name in names:
                if name in given:
                    new_args[name] = getattr(dt, name)
                else:
                    # If `name` is less significant than the most significant
                    # user-given value, fill in value from current time,
                    # otherwise fill in the lowest possible value from default.
                    # Example: User gave month and day: Get year from dt_now.
                    #          Get hours, minutes and seconds from dt_default.
                    if most_significant_given > names.index(name):
                        new_args[name] = getattr(dt_now, name)
                    else:
                        new_args[name] = getattr(dt_default, name)
            return datetime.datetime(**new_args)

        def get_timerange(dt, given):
            # Return min/max timestamp based on the least significant value.
            # Example: Use gave '2012-08' (year and month): Return timestamp for
            #          '2012-08-01 00:00:00' and '2012-08-31 23:59:59'.
            least_significant_given = given[-1]
            if least_significant_given == 'year':
                ts_min = datetime.datetime(year=dt.year, month=1, day=1)
                ts_max = datetime.datetime(year=dt.year + 1, month=1, day=1) - datetime.timedelta(seconds=1)
            elif least_significant_given == 'month':
                ts_min = datetime.datetime(year=dt.year, month=dt.month, day=1)
                if dt.month < 12:
                    ts_max = (datetime.datetime(year=dt.year, month=dt.month + 1, day=1)
                              - datetime.timedelta(seconds=1))
                else:
                    ts_max = (datetime.datetime(year=dt.year + 1, month=1, day=1)
                              - datetime.timedelta(seconds=1))
            elif least_significant_given == 'day':
                ts_min = datetime.datetime(year=dt.year, month=dt.month, day=dt.day)
                ts_max = ts_min + datetime.timedelta(hours=23, minutes=59, seconds=59)
            elif least_significant_given == 'hour':
                ts_min = datetime.datetime(year=dt.year, month=dt.month, day=dt.day, hour=dt.hour)
                ts_max = ts_min + datetime.timedelta(minutes=59, seconds=59)
            elif least_significant_given == 'minute':
                ts_min = datetime.datetime(year=dt.year, month=dt.month, day=dt.day, hour=dt.hour, minute=dt.minute)
                ts_max = ts_min + datetime.timedelta(seconds=59)
            else:
                ts_min, ts_max = dt, dt
            return (int(ts_min.timestamp()), int(ts_max.timestamp()))

        for frmt,given in cls._FORMATS:
            try:
                dt = datetime.datetime.strptime(string, frmt)
            except ValueError:
                continue
            else:
                dt = fill_in_missing_values(dt, given)
                return cls(dt.timestamp(), _timerange=get_timerange(dt, given))

        raise ValueError('Invalid format: %r' % string)

    def __new__(cls, seconds, _timerange=None):
        obj = super().__new__(cls, seconds)
        obj._timerange = _timerange if _timerange is not None else (seconds, seconds)
        return obj

    def __eq__(self, other):
        return self._timerange[0] <= other <= self._timerange[1]

    def __hash__(self):  # Needed to stay hashable with __eq__()
        return super().__hash__()

    def __gt__(self, other):
        return self._timerange[0] > other

    def __ge__(self, other):
        return self._timerange[1] >= other

    def __lt__(self, other):
        return self._timerange[1] < other

    def __le__(self, other):
        return self._timerange[0] <= other

    _SHORT_STR_MAP = {NOW: 'now', SOON: 'soon', UNKNOWN: '?',
                      NOT_APPLICABLE: '', NEVER: 'never'}

    def __str__(self):
        const_str = self._SHORT_STR_MAP.get(self, None)
        if const_str is not None:
            return const_str
        else:
            # The format is based on the delta of seconds compared to the
            # current time
            abs_delta = abs(self - time.time())
            if abs_delta < 120:
                frmt = '%H:%M:%S'
            elif abs_delta < 86400:
                frmt = '%H:%M'
            else:
                frmt = '%Y-%m-%d'
            return time.strftime(frmt, time.localtime(self))

    _LONG_STR_MAP = {NOW: 'now', SOON: 'soon', UNKNOWN: 'unknown',
                     NOT_APPLICABLE: 'not applicable', NEVER: 'never'}

    @cached_property
    def full(self):
        const_str = self._LONG_STR_MAP.get(self, None)
        if const_str is not None:
            return const_str
        else:
            return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(self))

    @cached_property
    def date(self):
        if self in self.CONSTANTS:
            return self.full
        else:
            return time.strftime('%Y-%m-%d', time.localtime(self))

    @cached_property
    def time(self):
        if self in self.CONSTANTS:
            return self.full
        else:
            return time.strftime('%H:%M:%S', time.localtime(self))

    def __bool__(self):
        return self not in (self.UNKNOWN, self.NOT_APPLICABLE, self.NEVER)

    @cached_property
    def is_known(self):
        return self not in self.CONSTANTS

    _DELTA_MAP = {NOW: Timedelta(0),
                  SOON: Timedelta(0),
                  UNKNOWN: Timedelta(Timedelta.UNKNOWN),
                  NOT_APPLICABLE: Timedelta(Timedelta.NOT_APPLICABLE),
                  NEVER: Timedelta(Timedelta.NOT_APPLICABLE)}

    @property
    def timedelta(self):
        delta = self._DELTA_MAP.get(self, None)
        if delta is not None:
            return delta
        else:
            return Timedelta(self - time.time())

    @property
    def in_future(self):
        return bool(self) and self > time.time()

    def __repr__(self):
        return '<%s %s / %s>' % (type(self).__name__, str(int(self)), self.full)


class PerfectInterval():
    """Remove processing time from intervals"""

    def __init__(self):
        self._last_timestamp = 0

    def __call__(self, seconds):
        now = asyncio.get_event_loop().time()
        if self._last_timestamp <= 0:
            self._last_timestamp = int(now)
            return seconds
        else:
            expected = self._last_timestamp + seconds
            diff = now - expected
            interval = max(seconds - diff, 0)
            self._last_timestamp = expected
            return interval


class SleepUneasy():
    """Asynchronous sleep() that can be aborted"""

    def __init__(self):
        self._interrupt = asyncio.Event()
        self._perfint = PerfectInterval()

    async def sleep(self, seconds):
        """Sleep for `seconds` or until `interrupt` is called"""
        self._interrupt.clear()
        # Remove processing time from seconds
        seconds = self._perfint(seconds)
        try:
            async with async_timeout(seconds):
                await self._interrupt.wait()
        except asyncio.TimeoutError:
            pass  # Interval passed without interrupt
        finally:
            self._interrupt.clear()

    def interrupt(self):
        """Stop sleeping"""
        self._interrupt.set()


class Response(SimpleNamespace):
    """
    Response to an API call

    All API implementations should use this class to provide return values to
    API calls.

    success: Whether the call was a success
    msgs: Sequence of messages; either strings or ClientError exceptions

    Any other keyword arguments are made available as attributes.
    """
    def __init__(self, success, msgs=(), errors=(), **kwargs):
        super().__init__(success=bool(success),
                         msgs=tuple(msgs), errors=tuple(errors),
                         **kwargs)


class LazyDict(dict):
    """Dictionary with callables as values that return the actual value on demand"""
    def __getitem__(self, key):
        value = dict.__getitem__(self, key)
        if callable(value):
            value = value()
            dict.__setitem__(self, key, value)
        return value


class URL():
    """Parse URL as lenient as possible (no validation)"""

    # Pilfered from yurl: https://github.com/homm/yurl
    # This is not validating regexp.
    # It splits url to unambiguous parts according RFC.
    _split_re = re.compile(r'''
        (?:([^:/?#]+):)?            # scheme
        (?://                       # authority
            (?:([^/?\#@]*)@)?       # userinfo
            ([^/?\#]*)              # host:port
        )?
        ([^?\#]*)                   # path
        \??([^\#]*)                 # query
        \#?(.*)                     # fragment
        ''', re.VERBOSE | re.DOTALL).match

    @classmethod
    def _parse_url(cls, url):
        # Default to http scheme so 'host:123' is not interpreted as 'host://123/'
        if len(url) > 0 and '://' not in str(url):
            url = 'http://' + str(url)

        groups = cls._split_re(url).groups('')
        # Order: scheme, user+password, host+port, path, query, fragment
        dct = {}
        dct['scheme']   = groups[0] or None
        dct['path']     = groups[3] or None
        dct['query']    = groups[4] or None
        dct['fragment'] = groups[5] or None

        auth = groups[1]
        user, pw = None, None
        if auth:
            # Password in authentication is optional
            if ':' in auth:
                user, pw = auth.split(':')
            else:
                user, pw = auth, None
        dct['user'] = user or None
        dct['password'] = pw or None

        # Split port from host only if it consists of digits.
        host = groups[2]
        port = None
        port_idx = host.rfind(':')
        if port_idx >= 0:
            port_str = host[port_idx + 1:]
            if port_str.isdigit():
                host = host[:port_idx]
                port = int(port_str)
        dct['host'] = host or None
        dct['port'] = port or None

        return dct

    _obj_cache = {}

    def __new__(cls, url):
        if isinstance(url, cls):
            return url

        cache_id = url
        cache = cls._obj_cache
        url_dict = cache.get(cache_id)
        if url_dict is None:
            url_dict = cache[cache_id] = cls._parse_url(url)

        obj = super().__new__(cls)
        for attr in ('scheme', 'host', 'port', 'path', 'user', 'password'):
            setattr(obj, attr, url_dict[attr])
        return obj

    @property
    def has_auth(self):
        """Whether user and password properties are set"""
        return self.user is not None and self.password is not None

    @property
    def domain(self):
        domain = getattr(self, '_domain_cached', None)
        if domain is None:
            host = self.host
            if not host:
                self._domain_cached = None
            else:
                if host.count('.') <= 1:
                    domain = self._domain_cached = host
                else:
                    parts = host.rsplit('.', maxsplit=2)
                    domain = self._domain_cached = '.'.join(parts[-2:])
        return domain

    def __str__(self):
        string = getattr(self, '_cached_string', None)
        if string is None:
            parts = []
            if self.scheme:
                parts.append('%s://' % self.scheme)
            if self.user:
                parts.append('%s' % self.user)
                if self.password:
                    parts.append(':%s' % self.password)
                parts.append('@')
            if self.host:
                parts.append(self.host)
            if self.port:
                parts.append(':%s' % self.port)
            if self.path:
                parts.append('%s' % self.path)
            string = self._cached_string = ''.join(parts)
        return string

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name[0] != '_':
            # Clear caches
            if hasattr(self, '_cached_string'):
                del self._cached_string
            if name == 'host' and hasattr(self, '_domain_cached'):
                del self._domain_cached

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, str(self))

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __le__(self, other):
        return str(self) <= str(other)

    def __gt__(self, other):
        return str(self) > str(other)

    def __ge__(self, other):
        return str(self) >= str(other)

    def __contains__(self, other):
        return str(other) in str(self)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Getters for tab completion candidates"""

import functools
import itertools
import os
import re
from collections import abc

from .. import objects
from ..completion import Candidate, Candidates
from ..utils import usertypes
from . import _utils

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


def for_args(args):
    """Get completion candidates for command line `args`"""
    cmdcls = objects.cmdmgr.get_cmdcls(args[0])
    if cmdcls is not None:
        log.debug('Getting candidates for args: %r', args)
        return cmdcls.completion_candidates(args)


@functools.lru_cache(maxsize=None)
def commands():
    """Names of commands"""
    cands = (Candidate(cmdcls.name,
                       in_parens='%s' % (', '.join(cmdcls.aliases),),
                       Description=cmdcls.description)
             for cmdcls in objects.cmdmgr.active_commands)
    return Candidates(cands, label='Command')


@functools.lru_cache(maxsize=None)
def help_topics():
    """All known help topics"""
    cats = []
    cats.append(Candidates(
        (Candidate(topic, Description=objects.helpmgr.MAIN_TOPICS[topic])
         for topic in objects.helpmgr.MAIN_TOPICS),
        label='Main Topic'))
    cats.append(commands())
    cats.append(Candidates(
        (Candidate(topic, Description=objects.localcfg.description(topic))
         for topic in objects.localcfg),
        label='Local Setting'))
    cats.append(Candidates(
        (Candidate(topic, Description=objects.remotecfg.description(topic))
         for topic in objects.remotecfg),
        label='Remote Setting'))
    return cats


@functools.lru_cache(maxsize=None)
def setting_names():
    """Names of settings"""
    return (Candidates((Candidate(name,
                                  Description=objects.localcfg.description(name),
                                  Default=str(objects.localcfg.default(name)))
                        for name in objects.localcfg),
                       label='Local Setting'),
            Candidates((Candidate(name,
                                  Description=objects.remotecfg.description(name),
                                  Default=str(objects.remotecfg.default(name)))
                        for name in objects.remotecfg),
                       label='Remote Setting'))

def setting_values(args):
    """
    Values of settings

    `args` must be a `stig.utils.cliparser.Args` instance in which the first
    argument is the name of a setting.
    """
    setting = args[0]
    try:
        value = objects.cfg[setting]
    except KeyError:
        return

    # Some settings accept multiple values, others only one
    focus_on_first_value = args.curarg_index == 1

    log.debug('Setting %r is a %s: %r', setting, type(value).__name__, value)
    # Get candidates depending on what kind of setting it is (bool, option, etc)
    if isinstance(value, usertypes.Option) and focus_on_first_value:
        aliases = value.aliases_inverse
        cands = (Candidate(opt, in_parens=', '.join(aliases.get(opt, '')))
                 for opt in value.options)
        return Candidates(cands, label='%s option' % (setting,))
    elif isinstance(value, usertypes.Tuple):
        aliases = value.aliases_inverse
        cands = (Candidate(opt, in_parens=', '.join(aliases.get(opt, '')))
                 for opt in value.options)
        return Candidates(cands, label='%s option' % (setting,),
                          curarg_seps=(value.sep.strip(),))
    elif isinstance(value, usertypes.Bool) and focus_on_first_value:
        options = (val
                   for vals in zip(value.truths, value.falsities)
                   for val in vals)
        return Candidates(options, label='%s option' % (setting,))
    elif isinstance(value, usertypes.Path):
        return fs_path(args.curarg.before_cursor,
                       base=value.base_path,
                       directories_only=os.path.isdir(value))


def sort_orders(clsname):
    """
    Names of sort orders

    `clsname` must be the name of a SorterBase derivative.
    """
    cls = _utils.get_sorter_cls(clsname)
    cands = (Candidate(name,
                       in_parens=', '.join(spec.aliases),
                       Description=spec.description)
             for name,spec in cls.SORTSPECS.items())
    settingname = 'sort.' + clsname[:-len('Sorter')].lower() + 's'
    return Candidates(cands,
                      curarg_seps=(objects.cfg[settingname].sep.strip(),),
                      label=_utils.sorters_labels[clsname])


def column_names(list_type):
    """
    Columns names for a certain type of list

    When `list_type` is appended to "columns.", it must resolve to a proper
    setting.
    """
    setting = objects.cfg['columns.%s' % list_type]
    cands = (Candidate(colname,
                       in_parens=', '.join(setting.aliases_inverse.get(colname, '')))
             for colname in setting.options)
    return Candidates(cands,
                      curarg_seps=(setting.sep.strip(),),
                      label=_utils.columns_labels[list_type])


def tab_titles():
    """Titles (strings) of TUI tabs"""
    from ..tui.tuiobjects import tabs
    return Candidates((widget.original_widget.text
                       for widget in tabs.titles), label='Tab Title')


def keybinding_contexts():
    """Arguments for the '--context' option of the 'bind' command"""
    from ..tui.tuiobjects import keymap
    return Candidates(keymap.contexts, label='Keybinding Context')

def keybinding_keys(args):
    """Bound keys"""
    from ..tui.tuiobjects import keymap
    keys = ()
    params = args.params(('--context', '-c'))
    if params:
        context = params[0]
        label = 'Keybinding in %s' % (context,)
        try:
            keys = tuple(keymap.keys(context=context))
        except ValueError:
            pass
        else:
            keys = (Candidate(key, Description=keymap.get_description(key, context=context))
                    for key in keys)
    else:
        label = 'Keybinding'
    return Candidates(keys, label=label)


def fs_path(path, base='.', directories_only=False, expand_home_directory=True, glob=None, regex=None):
    """
    File system path entries

    path: Path to get entries from
    base: Absolute path that is prepended to `path` if `path` is not absolute
    directories_only: Whether to include only directories
    expand_home_directory: Whether to interpret "~"
    glob: Unix shell pattern
    regex: Regular expression that is matched against each name in `path`
    """
    log.debug('Getting path candidates for %r', path)
    if path.startswith('~') and os.sep not in path and expand_home_directory:
        # Complete home dirs in "~<user>" style
        import pwd
        users = pwd.getpwall()
        cands = ('~%s' % (user.pw_name,) for user in users)
        label = 'Home directories'
    else:
        include_hidden = os.path.basename(path).startswith('.')
        if expand_home_directory:
            dirpath = os.path.expanduser(path)
        else:
            dirpath = path
        if not os.path.isabs(dirpath):
            dirpath = os.path.join(base, dirpath)
        dirpath = os.path.dirname(dirpath)
        try:
            itr = os.scandir(dirpath)
        except (FileNotFoundError, NotADirectoryError, PermissionError):
            cands = ()
            label = 'No matches'
        else:
            from fnmatch import fnmatch
            cands = (entry.name for entry in itr
                     if ((include_hidden or not entry.name.startswith('.')) and
                         (not directories_only or entry.is_dir()) and
                         (regex is None or entry.is_dir() or re.search(regex, entry.name)) and
                         (glob is None or entry.is_dir() or fnmatch(entry.name, glob))))
            if directories_only:
                label = 'Directories in %s' % (dirpath,)
            else:
                label = '%s' % (dirpath,)
            if glob:
                label += '/%s' % (glob,)
    return Candidates(cands, label=label, curarg_seps=('/',))


async def torrent_path(curarg, only='auto'):
    """
    If `curarg` is a path to a file or directory in a torrent ("<TORRENT
    FILTER>/<PATH>"), return one `Candidates` object for each matching torrent
    that has a matching path.  Otherwise, simply pass `curarg` to
    `torrent_filter` and return the result.

    `only` must be "files", "directories", "any" or "auto".  If `only` is
    "auto", candidates are taken from the parent directory; if `curarg` points
    to a file, only files are returned, if `curarg` points to a directory, only
    directories are returned.
    """
    parts = curarg.before_cursor.separate(('/',), include_seps=False)
    if len(parts) < 2:
        # User is still writing the torrent filter part
        log.debug('Passing %r to torrent_filter()', parts[0])
        # If the command line looks like 'ls comment=|/some/path' ('|' is the
        # cursor), '/' must be an argument separator to inform the completer
        # that '/some/path' is not the part we are completing.
        cats = await torrent_filter(parts[0])
        for cands in cats:
            cands.curarg_seps += ('/',)
        return cats

    def level_up(path):
        # Remove all empty parts from the right
        while len(path) > 1 and path[-1] == '':
            path = path[:-1]
        # Remove last part from the right
        return path[:-1] if path else path

    def get_cands(torrent, path, only):
        subtree = _utils.find_subtree(torrent, path)

        if subtree is not None:
            path_points_to_file = subtree.nodetype == 'leaf'
            if path_points_to_file:
                subtree = _utils.find_subtree(torrent, level_up(path))

            if only == 'files':
                cands = _utils.find_files(subtree)
                return Candidates(cands, curarg_seps=('/',), label='File in %s' % (subtree.path,))

            elif only == 'directories':
                cands = _utils.find_dirs(subtree)
                return Candidates(cands, curarg_seps=('/',), label='Directory in %s' % (subtree.path,))

            elif only == 'any':
                return Candidates(subtree, curarg_seps=('/',), label=subtree.path)

            elif only == 'auto':
                if path_points_to_file:
                    return get_cands(torrent, level_up(path), 'files')
                else:
                    return get_cands(torrent, level_up(path), 'directories')

            else:
                raise TypeError('Invalid value for argument "only": %r' % (only,))

    tfilter = parts[0]
    path = parts[1:]
    log.debug('Completing files or directories in %r torrents: %r', tfilter, path)
    response = await objects.srvapi.torrent.torrents(tfilter, keys=('files',), from_cache=True)
    if response.success:
        return (get_cands(torrent, path, only)
                for torrent in response.torrents)


async def torrent_filter(curarg, filter_names=True):
    """
    Values and/or names for torrent filters

    If `filter_names` evaluates to False, filter names are not included in the
    returned list, only torrent names.

    The return value is either an empty tuple, a 1-tuple (filter values) or a
    2-tuple (filter names and filter values).
    """
    async def objects_getter(keys):
        response = await objects.srvapi.torrent.torrents(keys=keys, from_cache=True)
        if response.success:
            return response.torrents
    return await _filter(curarg, 'TorrentFilter',
                         objects_getter=objects_getter,
                         items_getter=None,
                         filter_names=filter_names)

async def file_filter(curarg, torrent_filter, filter_names=True):
    """Values and/or names for file filters (see `torrent_filter`)"""
    async def objects_getter(keys):
        response = await objects.srvapi.torrent.torrents(torrent_filter, keys=keys, from_cache=True)
        if response.success:
            return response.torrents

    # Get list of files from Torrent instance
    def items_getter(t):
        # Exclude single-file torrents (torrents that don't contain a
        # directory); file filters are irrelevant in that case.
        files = tuple(t['files'].files)
        if len(files) != 1 or t['name'] != files[0]['name']:
            return files
        else:
            return ()
    return await _filter(curarg, 'FileFilter',
                         objects_getter=objects_getter,
                         items_getter=items_getter,
                         filter_names=filter_names)

async def peer_filter(curarg, torrent_filter, filter_names=True):
    """Values and/or names for peer filters (see `torrent_filter`)"""
    async def objects_getter(keys):
        response = await objects.srvapi.torrent.torrents(torrent_filter, keys=keys, from_cache=True)
        if response.success:
            return response.torrents
    return await _filter(curarg, 'PeerFilter',
                         objects_getter=objects_getter,
                         # Get list of peers from Torrent instance
                         items_getter=lambda t: t['peers'],
                         filter_names=filter_names)

async def tracker_filter(curarg, torrent_filter, filter_names=True):
    """Values and/or names for tracker filters (see `torrent_filter`)"""
    async def objects_getter(keys):
        response = await objects.srvapi.torrent.torrents(torrent_filter, keys=keys, from_cache=True)
        if response.success:
            return response.torrents
    return await _filter(curarg, 'TrackerFilter',
                         objects_getter=objects_getter,
                         # Get list of trackers from Torrent instance
                         items_getter=lambda t: t['trackers'],
                         filter_names=filter_names)

async def setting_filter(curarg, filter_names=True):
    """Values and/or names for setting filters (see `torrent_filter`)"""
    async def objects_getter(**_):
        return objects.cfg.as_dict.values()
    return await _filter(curarg, 'SettingFilter',
                         objects_getter=objects_getter,
                         items_getter=None,
                         filter_names=filter_names)
async def labels(curarg):
    """All labels"""
    labels = set()
    if curarg != "":
        response = await objects.srvapi.torrent.torrents(None, ('labels', ), from_cache=True)
        [
            [labels.add(l) for l in t["labels"] if l.startswith(curarg)]
            for t in response.torrents
        ]
    return Candidates(list(labels), label=curarg, curarg_seps=',')


async def _filter(curarg, filter_cls_name, objects_getter, items_getter, filter_names):
    """
    Values and/or names for filters

    `objects_getter` is a coroutine function that returns a sequence of objects
    that are used to get values via the filter's value_getter (e.g. Torrent).

    If `items_getter` is not None, it is a callable that gets an object (see
    above) and returns a list of items that are used to get values via the
    filter's value_getter (e.g. TorrentPeer or TorrentTracker).  If
    `items_getter` is None, the objects returned by `objects_getter` are used as
    items.

    If `filter_names` evaluates to False, filter names are not included in the
    returned list, i.e. only values for the default filter (e.g. torrent name)
    are completed.

    The return value is either an empty tuple, a 1-tuple (filter values) or a
    2-tuple (filter names and filter values).
    """
    filter_cls = _utils.get_filter_cls(filter_cls_name)
    default_filter = filter_cls.DEFAULT_FILTER

    # Remove any leading special character from the filter
    if curarg.startswith(filter_cls.INVERT_CHAR):
        curarg = curarg[1:]
        # If the cursor was on a special character, move it to the first
        # character of the filter name
        if curarg.curpos is None:
            curarg.curpos = 0

    # Separate individual filters, e.g. 'seeding|comment=foo' -> ('seeding', 'comment=foo')
    filter_strings = curarg.separate(_utils.filter_combine_ops, include_seps=True)

    # Most arguments to _filter_values() are the same
    filter_values = functools.partial(_filter_values, filter_cls_name,
                                      objects_getter=objects_getter,
                                      items_getter=items_getter)

    # Separate filter name from filter value
    parts = filter_strings.curarg.separate(_utils.filter_compare_ops, include_seps=True)
    if parts.curarg_index == 0:
        # If focus is on filter name, complete filter names and values for the
        # default filter (usually "name")
        if filter_names:
            log.debug('Completing filter names and %r values: %r', default_filter, parts[0])
            return (_utils.filter_names(filter_cls_name),
                    await filter_values(default_filter))
        else:
            log.debug('Completing %r values (no filter names): %r', default_filter, parts[0])
            return (await filter_values(default_filter),)

    elif parts.curarg_index == 1 and parts[0] in _utils.filter_compare_ops:
        # User gave a comparison operator but not a filter name, e.g. ('=', 'foo')
        log.debug('Completing %r (default) values', default_filter)
        return (await filter_values(default_filter),)

    elif parts.curarg_index == 2:
        # User gave filter name, operator and value, e.g. ('comment', '!=', 'foo')
        filter_name = parts[0].strip()
        log.debug('Completing %r values', filter_name)
        return (await filter_values(filter_name),)
    else:
        return ()

async def _filter_values(filter_cls_name, filter_name, objects_getter, items_getter):
    # Return list of possible values for filter (if filter is comparative)
    filter_cls = _utils.get_filter_cls(filter_cls_name)
    cands = ()
    if _utils.filter_takes_completable_values(filter_cls, filter_name):
        filter_spec = _utils.get_filter_spec(filter_cls, filter_name)

        # Some filters (e.g. file filter "priority") have a fixed set of allowed values).
        if hasattr(filter_spec.value_type, 'valid_values'):
            cands = filter_spec.value_type.valid_values

        # Get a list of items (torrents, files, peers, etc) and ask the filter
        # class for the values it would match against.
        else:
            keys = filter_cls(filter_name).needed_keys
            objects = await objects_getter(keys=keys)
            if objects:
                cands = []

                def add_cands_for(item):
                    # Get the same value from `item` that the filter would get
                    value_getter = filter_spec.value_getter
                    value = value_getter(item)
                    # Some value_getters return multiple values, e.q. the torrent
                    # filter "tracker", which matches against domain names of
                    # announce URLs.
                    if isinstance(value, (abc.Iterable, abc.Iterator)) and not isinstance(value, str):
                        value = tuple(value)
                        cands.extend(value)
                    else:
                        cands.append(value)

                # File
                if items_getter is None:
                    # Filtering objects directly (e.g. Torrent instances)
                    for obj in objects:
                        add_cands_for(obj)
                else:
                    # Filtering items from a sub-list of an object,
                    # (e.g. torrent['files'] or torrent['peers'])
                    for obj in objects:
                        items = items_getter(obj)
                        for i in items:
                            add_cands_for(i)

    curarg_seps = itertools.chain(_utils.filter_compare_ops, _utils.filter_combine_ops)
    label = '%s: %s' % (_utils.filter_labels[filter_cls_name], filter_name)
    return Candidates(cands, label=label, curarg_seps=curarg_seps)
